/*
 *	hash.c: part of oar(Object ARchiver)
 *	Copyright  1995  씎Y(oo), HGG00736@niftyserve.or.jp
 */
#include	"oar.h"
#include	<string.h>
#ifdef	HAVE_HMEMORY_H
#include	"hmemory.h"
#endif


#ifndef	HASHTAB_SIZE
#define	HASHTAB_SIZE	(0x1000)
#endif	/*HASHTAB_SIZE*/
#define	HASHTAB_MASK		(HASHTAB_SIZE-1)

#ifndef	HASH_RETRYMAX
#define	HASH_RETRYMAX	8
#endif	/*HASH_RETRYMAX*/

#ifdef	DEBUG
#define	STATIC
#else
#define	STATIC	static
#endif

STATIC int	hash_count = 0;
STATIC int	hash_conflict = 0;
STATIC hash_s*	hashtab = (hash_s*)0;
STATIC unsigned	hashtab_size = HASHTAB_SIZE;
STATIC unsigned	hashtab_mask = HASHTAB_MASK;

int	(*hash_strcmp)(const char*, const char*) = stricmp;


#ifdef	__GNUC__
#define	/*unsigned*/hash(/*name_t*/s,mask)	\
({\
	name_t _l_s_ = (s);\
	unsigned _l_c_;\
	unsigned _l_val_ = 0;\
	while (_l_c_ = *_l_s_++) {\
		if (ignore_cases) {\
			if ((byte)(_l_c_ - 'A') <= (byte)('Z' - 'A')) _l_c_ += ('a' - 'A');\
		}\
		_l_val_ = _l_c_ + (_l_val_<<5) - _l_val_;\
	}\
	(unsigned)(_l_val_&(mask));\
})
#else
static unsigned hash(name_t s, unsigned mask)
{
	unsigned c;
	unsigned val = 0;

	while (c = *s++) {
		if (ignore_cases) {
			if ((byte)(c - 'A') <= (byte)('Z' - 'A')) c += ('a'-'A');
		}
		val = c + (val<<5) - val;
	}
	return val & mask;
}
#endif


int	hash_init(void)
{
#ifndef	HAVE_HMEMORY_H
	unsigned	i;
#endif

	if (hashtab != (hash_s*)0) return 0;

	hash_count = 0;
	if ((hashtab = (hash_s*)malloc(hashtab_size*sizeof(hash_s))) == (hash_s*)0) {
		return -1;
	}
#ifdef	HAVE_HMEMORY_H
	_blkset((long*)hashtab, (long)((name_t)0), (hashtab_size*sizeof(hash_s))/256);
#else	/*!HAVE_HMEMORY_H*/
	i = hashtab_size;
	while (i-- > 0) {
		hashtab[i].obj_name = (name_t)0;
	}
#endif
	return 0;
}


void	hash_term(void)
{
	if (hashtab) free(hashtab);
}


hash_s*	hash_resize(name_t name)
{
	int	retry;
	hash_s*	new_tab;
	unsigned new_size, new_mask;
	unsigned hashval, new_val;

	hash_conflict = 0;
	new_size = hashtab_size << 1;

	/*
	 *	SmۂĂ̂ŗ]
	 */
	if (new_size*sizeof(hash_s) + 0x10 >= archive_limit - archive_size) {
		return (hash_s*)0;
	}
	archive_limit -= new_size*sizeof(hash_s) + 0x10;
	if ((void*)realloc(archive_buff, archive_limit) == (void*)0) {
		return (hash_s*)0;
	}

	if (new_tab = (hash_s*)malloc(new_size*sizeof(hash_s))) {
#ifdef	DEBUG
		fprintf(stderr, "hash resize starting...\n");
#endif
		/*
		 *	initialize new table
		 */
#ifdef	HAVE_HMEMORY_H
		_blkset((void*)new_tab, (long)((name_t)0), new_size/(256/sizeof(hash_s)));
#else	/*HAVE_HMEMORY_H*/
		new_val  = new_size;
		while (new_val-- > 0) {
			new_tab[new_val].obj_name = (name_t)0;
		}
#endif
		new_mask = new_size - 1;

		for (hashval = 0; hashval < hashtab_size; hashval++) {
			if (hashtab[hashval].obj_name != (name_t)0) {
				retry = 0;
#ifdef	DEBUG
				fprintf(stderr, "\t%04x:%s\n", hashval, hashtab[hashval].obj_name);
#endif
				new_val = hash(hashtab[hashval].obj_name, new_mask);
				while (new_tab[new_val].obj_name != (name_t)0) {
					++retry;
					new_val = ++new_val & new_mask;
				}
				if (retry) ++hash_conflict;
				new_tab[new_val].obj_name  = hashtab[hashval].obj_name;
				new_tab[new_val].name_ptr = hashtab[hashval].name_ptr;
#ifndef	DEBUG
				(new_tab[new_val].name_ptr)->hash_ptr = &new_tab[new_val];
#endif
			}
		}
		free(hashtab);
		hashtab = new_tab;
		hashtab_size = new_size;
		hashtab_mask = new_mask;
#ifdef	DEBUG
		fprintf(stderr, "done\n");
#endif

		retry = 0;
		hashval = hash(name, hashtab_mask);
		while (hashtab[hashval].obj_name != (name_t)0) {
			if (++retry > HASH_RETRYMAX) return hash_resize(name);
			hashval = ++hashval & hashtab_mask;
		}
		if (retry) {
			++hash_conflict;
#ifdef	DEBUG
			fprintf(stderr, "conflict %d(%s)\n", retry, name);
#endif
		}
		++hash_count;
		hashtab[hashval].obj_name = name;
		return &hashtab[hashval];
	}
	return (hash_s*)0;
}


hash_s*	hash_install(name_t name)
{
	int	retry = 0;
	unsigned hashval;

	hashval = hash(name, hashtab_mask);
	while (hashtab[hashval].obj_name != (name_t)0) {
		if (++retry > HASH_RETRYMAX) return hash_resize(name);
		hashval = ++hashval & hashtab_mask;
	}
	if (retry) {
		++hash_conflict;
#ifdef	DEBUG
		fprintf(stderr, "conflict %d(%s)\n", retry, name);
#endif
	}
	++hash_count;
	hashtab[hashval].obj_name = name;
	return &hashtab[hashval];
}


hash_s*	hash_lookup(name_t name)
{
	int	retry = 0;
	unsigned hashval;
	name_t	hashname;

	hashval = hash(name, hashtab_mask);
	while ((hashname = hashtab[hashval].obj_name) != (name_t)0) {
		if (++retry > HASH_RETRYMAX) return (hash_s*)0;
		if ((*hash_strcmp)(hashname, name) == 0) return &hashtab[hashval];
		hashval = ++hashval & hashtab_mask;
	}
	return (hash_s*)0;
}


void	hash_dump(FILE* fp)
{
	name_t	name;
	unsigned hashval;

	if (hash_count) {
		for (hashval = 0; hashval < hashtab_size; hashval++) {
			if ((name = hashtab[hashval].obj_name) != (name_t)0) {
				fprintf(fp, "%04x:%s\n", hashval, name);
			}
		}
	}
	fprintf(fp, "%5d/%5d(%3d%%), conflict=%d(%3d%%)\n",
			hash_count, hashtab_size, (hash_count*100)/hashtab_size,
			hash_conflict, (hash_conflict*100)/hash_count);
}


#ifdef	DEBUG
#include	"readln.h"

int	main(int ac, char** av)
{
	name_t	name;
	FILE*	fp;

	if (ac != 2) {
		fprintf(stderr, "%s file\n", av[0]);
		return -1;
	}
	hash_init();
	if (fp = fopen(av[1], "rb")) {
		while (name = readln_b(fp)) {
			if ((name = (name_t)strdup(name)) == (name_t)0) {
				fprintf(stderr, "strdup error\n");
			}
			if (hash_install(name) == (hash_s*)0) {
				fprintf(stderr, "name conflict %s(%d)\n", name, hash_count);
			}
		}
	}
	hash_dump(stdout);
	hash_term();
	return 0;
}
#endif	/*DEBUG*/
