/*
 *	oar.c: part of oar(Object ARchiver)
 *	Copyright  1995  씎Y(oo), HGG00736@niftyserve.or.jp
 */
#include	"oar.h"
#include	<limits.h>
#include	"getopt.h"
#include	"error.h"
#include	"strmacro.h"


VOLATILE void	oar_term(int code)
{
	name_term();
	hash_term();
	if (archive_buff) free(archive_buff);
	if (temp_archive) {
		remove(temp_archive);
		temp_archive = (char*)0;
	}
	exit(code);
}


static void	oar_abort(void)
{
	if (temp_archive) remove(temp_archive);
}


static void	print_version(void)
{
	fprintf(stderr, "oar ver" VERSION " Copyright 1995 oo\n");
}

/*
 *	option variables for GNU getopt()
 */
int		help = 0;
int		force = 1;
int		verbose = 0;
int		version = 0;
int		chk_null = 1;
int		long_form = 0;
int		all_label = 0;
int		concatenate = 0;
int		ignore_cases = 1;

struct option long_opt[] = {
	{"all-label",		no_argument,		NULL,		'A'},
	{"concatenate", 	no_argument,		NULL,		'C'},
	{"decimal", 		no_argument,		NULL,		'D'},
	{"force",			no_argument,		NULL,		'F'},
	{"get-libpath", 	no_argument,		NULL,		'G'},
	{"long-format", 	no_argument,		NULL,		'L'},
	{"no-check_null",	no_argument,		NULL,		'N'},
	{"case-sensitive",	no_argument,		NULL,		'S'},
	{"update",			no_argument,		NULL,		'U'},
	{"version", 		no_argument,		NULL,		'V'},
	{"hexadecimal", 	no_argument,		NULL,		'X'},
	{"append",			no_argument,		NULL,		'a'},
	{"create",			no_argument,		NULL,		'c'},
	{"delete",			no_argument,		NULL,		'd'},
	{"extract-all", 	no_argument,		NULL,		'e'},
	{"help",			no_argument,		NULL,		'h'},
	{"list",			no_argument,		NULL,		'l'},
	{"merge",			no_argument,		NULL,		'm'},
	{"print-label", 	no_argument,		NULL,		'p'},
	{"quiet",			no_argument,		NULL,		'q'},
	{"replace", 		no_argument,		NULL,		'r'},
	{"table",			no_argument,		NULL,		't'},
	{"force-append",	no_argument,		NULL,		'u'},
	{"verbose", 		no_argument,		NULL,		'v'},
	{"extract", 		no_argument,		NULL,		'x'},
	{0, 0, 0, 0}
};

struct option arlong_opt[] = {
	{"all-label",		no_argument,		NULL,		'A'},
	{"concatenate", 	no_argument,		NULL,		'C'},
	{"decimal", 		no_argument,		NULL,		'D'},
	{"force",			no_argument,		NULL,		'F'},
	{"get-libpath", 	no_argument,		NULL,		'G'},
	{"long-format", 	no_argument,		NULL,		'L'},
	{"no-check_null",	no_argument,		NULL,		'N'},
	{"case-sensitive",	no_argument,		NULL,		'S'},
	{"update",			no_argument,		NULL,		'U'},
	{"version", 		no_argument,		NULL,		'V'},
	{"hexadecimal", 	no_argument,		NULL,		'X'},
	{"quiet",			no_argument,		NULL,		'c'},
	{"delete",			no_argument,		NULL,		'd'},
	{"extract-all", 	no_argument,		NULL,		'e'},
	{"help",			no_argument,		NULL,		'h'},
	{"modify",			no_argument,		NULL,		'm'},
	{"print-label", 	no_argument,		NULL,		'p'},
	{"force-append",	no_argument,		NULL,		'q'},
	{"replace",			no_argument,		NULL,		'r'},
	{"table",			no_argument,		NULL,		't'},
	{"update",			no_argument,		NULL,		'u'},
	{"verbose",			no_argument,		NULL,		'v'},
	{"extract",			no_argument,		NULL,		'x'},
	{0, 0, 0, 0}
};

#define	GETOPT_STRING	"ACDFGLNSUVXabcdeghilmopqrstuvxz"
#define	OPTION_STRING	"[-ACDFGLNSUVXacdehlmpqrtuvx]"
#define	AROPTION_STRING	"[-]ACDFGLNSUVXcdehmpqrtuvx"

static void	print_help(void)
{
	fprintf(stderr,
#if	1
"gp@: %s %s A[JCut@C [IuWFNgt@C]\n"
"\t-A --all-label       A[JCûׂẴx̕\\n"
"\t-C --concatenate     IuWFNgXy[Xŋ؂ďo\n"
"\t-D --decimal         l\POitH[}bgɂ\n"
"\t-F --force            append/extract/replace s\n"
"\t-G --get-libpath     ϐ lib A[JCũpXɂ\n"
"\t-L --long-format     ڍׂȕ\`ɂ\n"
"\t-N --no-check-null   null foCX`FbN}\n"
"\t-S --case-sensitive  IuWFNg̑啶Eʂ\n"
"\t-U --update          VKt@C̎̂ append/extract/replace s\n"
"\t-V --version         o[W̕\\n"
"\t-X --hexadecimal     l\PUitH[}bgɂ\n",
#else
"usage: %s %s archive_file [object_file(s)]\n"
"\t-A --all-label       print all labels in an archive\n"
"\t-C --concatenate     concatenate all object names\n"
"\t-D --decimal         print value with decimal format\n"
"\t-F --force           force append/extract/replace(default)\n"
"\t-G --get-libpath     use environment variable `lib\' for archive file path\n"
"\t-L --long-format     print long format\n"
"\t-N --no-check-null   no checking null device\n"
"\t-S --case-sensitive  case sensitive of object name\n"
"\t-U --update          update mode\n"
"\t-V --version         print oar version number\n"
"\t-X --hexadecimal     print value with hexadecimal format\n",
#endif
	progname, (ar_mode ? AROPTION_STRING : OPTION_STRING));

	fputs(
#if	1
(ar_mode
?
"\t-c --quiet           [h\n"
"\t-d --delete          IuWFNg폜\n"
"\t-e --extract-all     SẴIuWFNg̒o\n"
"\t-h --help            ̃bZ[W̕\\n"
"\t-m --modify          IuWFNgǉ\n"
"\t-p --print-label     A[JCůO`EOQƃx̕\\n"
"\t-q --force-append    IuWFNgǉ\n"
"\t-r --replace         IuWFNgu(default)\n"
"\t-t --table           IuWFNg\\n"
"\t-u --update          VKt@C̎̂ append/extract/replace s\n"
"\t-v --verbose         ׂ胂[h\n"
"\t-x --extract         IuWFNg𒊏o\n"
:
"\t-a --append          IuWFNgǉ\n"
"\t-c --create          A[JCu̐VK쐬\n"
"\t-d --delete          IuWFNg폜\n"
"\t-e --extract-all     SẴIuWFNg̒o\n"
"\t-h --help            ̃bZ[W̕\\n"
"\t-l --list            IuWFNg\\n"
"\t-m --merge           ̃A[JCu̘A\n"
"\t-p --print-label     A[JCůO`EOQƃx̕\\n"
"\t-q --quiet           [h\n"
"\t-r --replace         IuWFNgu(default)\n"
"\t-t --table           IuWFNg\\n"
"\t-u --force-append    IuWFNgǉ\n"
"\t-v --verbose         ׂ胂[h\n"
"\t-x --extract         IuWFNg𒊏o\n"
),
#else
"\t-a --append          append files to end of an archive\n"
"\t-c --create          create a new archive\n"
"\t-d --delete          delete files from an archive\n"
"\t-e --extract-all     extract all files from an archive\n"
"\t-h --help            print this message\n"
"\t-l --list            list the contents of an archive\n"
"\t-l --merge           merge archives to the first archive\n"
"\t-p --print-label     print labels of all objects from an archive\n"
"\t-q --quiet           set quiet mode\n"
"\t-r --replace         replace files to an archive(default)\n"
"\t-t --table           list the contents of an archive\n"
"\t-v --verbose         set verbose mode\n"
"\t-x --extract         extract files from an archive\n",
#endif
	stderr);
	oar_term(-1);
}


int	ar_mode = 0;
char	archive_path[PATH_MAX + 1] = "";
char*	archive_file = (char*)0;
char*	archive_dir  = (char*)0;
int		(*function)(int, char**) = F_replace;
char*	(*format)(char*, long*)  = ltodazs;


void	oar_init(void)
{
	fatal_term = oar_term;
	if (hash_init() == -1 || name_init() == -1) {
		fatal("vȌɃG[܂");
	}
	onexit(oar_abort);
}


static void	options(int optch)
{
	switch (optch) {
	case 'A':
		all_label = 1;
		function = F_print;
		break;
	case 'C':
		concatenate = 1;
		break;
	case 'D':
		format = ltodazs;
		break;
	case 'F':
		force = 1;
		break;
	case 'G':
	case 'g':
		if (archive_dir == (char)0
		&&  (archive_dir = getenv("lib")) == (char*)0) {
			archive_dir = "";
			warning("ϐ `lib\' `Ă܂");
		}
		break;
	case 'L':
		long_form = 1;
		break;
	case 'N':
		chk_null = 0;
		break;
	case 'S':
		ignore_cases = 0;
		hash_strcmp = strcmp;
		break;
	case 'U':
		force = 0;
		break;
	case 'V':
		version = 1;
		break;
	case 'X':
		format = ltoha;
		break;
	case 'a':
		if (!ar_mode) function = F_append;
	case 'b':
		break;
	case 'c':
		if (ar_mode) {
			verbose = 0;
		}
		else {
			function = F_create;
		}
		break;
	case 'd':
		function = F_delete;
		break;
	case 'e':
		function = F_extall;
		break;
	case 'h':
		help = 1;
	case 'i':
		break;
	case 'l':
		if (!ar_mode) function = F_list;
		break;
	case 'm':
		if (ar_mode) {
			function = F_append;
		}
		else {
			function = F_merge;
		}
	case 'o':
		break;
	case 'p':
		function = F_print;
		break;
	case 'q':
		if (ar_mode) {
			force = 1;
			function = F_append;
		}
		else {
			verbose = 0;
		}
		break;
	case 'r':
		function = F_replace;
	case 's':
		break;
	case 't':
		function = F_list;
		break;
	case 'u':
		if (ar_mode) {
			force = 0;
		}
		else {
			force = 1;
			function = F_append;
		}
		break;
	case 'v':
		verbose = 1;
		break;
	case 'x':
		function = F_extract;
	case 'z':
		break;
	case '?':	/* unkown option */
	default:
		print_help();
		break;
	}
}


int	main(int argc, char** argv)
{
	int	c, count;
	char*	ptr;
	char*	ar_options = NULL;

	set_progname(argv[0]);
	if (stricmp(progname, "ar") == 0) {
		ar_mode = 1;
		if (argc == 1) exit(0);
	}
	oar_init();

	optind = 0;
	if (!ar_mode && (ptr = getenv("OAR"))) {
		while (c = *ptr++) {
			if ((unsigned char)c == (unsigned char)'\t'
			||  (unsigned char)c == (unsigned char)' '
			||  (unsigned char)c == (unsigned char)'-') continue;
			options(c);
		}
	}

	/* UNIX ar Ƃ̌݊̂ */
	if (ar_mode && argv[1][0] != '-') {
		if ((ar_options = (char*)malloc(strlen(argv[1]) + 2)) == NULL) {
			fatal("IvVWĴ߂̃mۂł܂");
		}
		*ar_options = '-';
		strcpy(ar_options + 1, argv[1]);
		argv[1] = ar_options;
	}

	optind = 0;
	while ((c = getopt_long(argc, argv, GETOPT_STRING,
	                        (ar_mode ? arlong_opt : long_opt), &count)) != EOF) {
		if (c == 0) {
			*long_opt[count].flag = 1;
		}
		else {
			options(c);
		}
	}
	if (ar_options) free(ar_options);

	if (version) {
		print_version();
		if (argc == 2
		&&  (strcmp(argv[1], "-V") == 0 || strcmp(argv[1], "--version") == 0)) oar_term(0);
	}
	if (argc == optind || help) print_help();

	/*
	 *	A[JCut@C̐ݒ
	 */
	if (archive_dir && *archive_dir != '\0') {
		ptr = strcpy(archive_path, archive_dir) + strlen(archive_dir) - 1;
		if ((char*)mbsrchr(archive_path, '\\') != ptr
		&&  (char*)mbsrchr(archive_path, '/')  != ptr) {
			*++ptr = '/';
		}
		strcpy(++ptr, archive_file = argv[optind++]);
	}
	else {
		strcpy(archive_path, archive_file = argv[optind++]);
	}
	toslash(archive_path);

	oar_term((*function)(argc - optind, argv + optind));
}


