#ifndef	_OAR_H_
/*
 *	oar.h: part of oar(object archiver)
 *	Copyright  1994  씎Y(oo), HGG00736@niftyserve.or.jp
 */
#define	_OAR_H_	1

#ifdef	HAVE_CONFIG_H
#include	"config.h"
#endif	/*HAVE_CONFIG_H*/
#include	<stdio.h>
#ifdef	HAVE_STDLIB_H
#include	<stdlib.h>
#endif	/*HAVE_STDLIB_H*/
#include	"ansidecl.h"
#ifdef	HAVE_IPRINT_H
#define	CALL_IPRINT_ALL
#include	"iprint.h"
#endif	/*HAVE_IPRINT_H*/

typedef	unsigned char*	name_t;
typedef	unsigned char	byte;
typedef	unsigned short	word;
typedef	unsigned long	lword;

#define	ARC_HEADER1	((lword)0xd1000000UL)
#define	ARC_HEADER2	((word)0x0002)
#define	ARC_HEADSIZ	6
#define	OBJNAME_MAX		0x18

typedef	struct {
	byte	obj_name[OBJNAME_MAX];
	lword	obj_size;
	word	obj_time;
	word	obj_date;
}	header_s;

#define	HEADERSIZ	sizeof(header_s)

extern header_s	header;


/*
 *	oar.c(main)
 */
extern int	force;
extern int	verbose;
extern int	chk_null;
extern int	long_form;
extern int	all_label;
extern int	concatenate;
extern int	ignore_cases;
extern int	ar_mode;

extern char		archive_path[];
extern char*	archive_file;
extern char*	object_file;

extern void				oar_init(void);
extern VOLATILE void	oar_term(int code);

extern char* (*format)(char*, long*);


/*
 *	hash.c
 */
typedef	struct hashtable	{
	name_t	obj_name;
	struct nametable*	name_ptr;
}	hash_s;

extern int	(*hash_strcmp)(const char*, const char*);

extern int		hash_init(void);
extern void		hash_term(void);
extern hash_s*	hash_resize(name_t name);
extern hash_s*	hash_install(name_t name);
extern hash_s*	hash_lookup(name_t name);
extern void		hash_dump(FILE* fp);


/*
 *	name.c
 */
typedef	struct nametable	{
	name_t	obj_name;
	struct hashtable*	hash_ptr;
}	name_s;

extern name_s*	object_names;
extern size_t	object_count;
extern size_t	object_limit;
extern size_t	delete_count;

extern int		name_init(void);
extern void		name_term(void);
extern int		name_resize(name_t name, hash_s* hashptr);
extern void		name_dump(void);

#define	/*name_t*/name_install(/*name_t*/n,/*hash_s**/hp)	\
		(object_count<object_limit\
		?object_names[object_count].obj_name=(n),\
		 object_names[object_count].hash_ptr=(hp),\
		 ((hash_s*)(hp))->name_ptr=object_names+object_count++,0\
		:name_resize((n),(hp)))

/*
 *	dosio.c
 */
extern int		old_archive;
extern char*	temp_archive;
extern byte*	archive_buff;
extern byte*	archive_ptr;
extern byte*	old_archive_ptr;
extern size_t	archive_size;
extern size_t	archive_limit;

extern byte* skip_dirs(byte* path);
extern void	check_null_device(void);
extern void	read_archive(int mode);
extern void	read_archive2(name_t name);
extern int	write_archive(void);
extern int	read_object(name_t name);


/*
 *	subroutines
 */
extern void	append_object(void);			/* append.c */
extern int	F_append(int ac, char** av);	/* append.c */
extern int	F_create(int ac, char** av);	/* create.c */
extern int	F_delete(int ac, char** av);	/* delete.c */
extern int	F_extract(int ac, char** av);	/* extract.c */
extern int	F_extall(int ac, char** av);	/* extract.c */
extern void	list_object(header_s* ptr);		/* list.c */
extern int	F_list(int ac, char** av);		/* list.c */
extern int	F_merge(int ac, char** av);		/* merge.c */
extern int	F_print(int ac, char** av);		/* print.c */
extern int	F_replace(int ac, char** av);	/* replace.c */


/*
 *	toslash.c
 */
extern char* toslash(char* path);


/*
 *	dosdate.s
 */
typedef	struct	{
	word	obj_time;
	word	obj_date;
}	dostime_s;

extern char* dosdate(char* dst, dostime_s* ptr);

/*
 *	ltoda.s
 */
extern char* ltoda(char* dst, long* ptr);
extern char* ltodazs(char* dst, long* ptr);

/*
 *	ltoha.s
 */
extern char* ltoha(char* dst, long* ptr);
extern char* ltohazs(char* dst, long* ptr);
extern char* ltohazs6(char* dst, long* ptr);


#endif	/*_OAR_H_*/
